@@ -1,13 +1,16 @@
 """XMPP bots for humans."""
 
-from argparse import ArgumentParser, BooleanOptionalAction
+from argparse import ArgumentParser
 from configparser import ConfigParser
+from datetime import datetime as dt
 from getpass import getpass
-from logging import DEBUG, INFO, basicConfig
+from logging import DEBUG, INFO, basicConfig, getLogger
 from os import environ
 from os.path import exists
 from pathlib import Path
+from sys import exit, stdout
 
+from humanize import naturaldelta
 from slixmpp import ClientXMPP
 
 
@@ -42,10 +45,36 @@ class SimpleMessage:
         return self.message["type"]
 
 
+class Config:
+    """Bot file configuration."""
+
+    def __init__(self, name, config):
+        self.name = name
+        self.config = config
+        self.section = config[self.name] if self.name in config else {}
+
+    @property
+    def account(self):
+        return self.section.get("account", None)
+
+    @property
+    def password(self):
+        return self.section.get("password", None)
+
+    @property
+    def nick(self):
+        return self.section.get("nick", None)
+
+
 class Bot(ClientXMPP):
-    CONFIG_FILE = "bot.conf"
+    """XMPP bots for humans."""
 
     def __init__(self):
+        self.name = type(self).__name__.lower()
+        self.start = dt.now()
+
+        self.CONFIG_FILE = f"{self.name}.conf"
+
         self.parse_arguments()
         self.setup_logging()
         self.read_config()
@@ -56,22 +85,36 @@ class Bot(ClientXMPP):
 
     def parse_arguments(self):
         """Parse command-line arguments."""
-        self.parser = ArgumentParser()
-        self.parser.add_argument(
-            "--input",
-            help="Read configuration from environment",
-            action=BooleanOptionalAction,
-            default=True,
-        )
+        self.parser = ArgumentParser(description="XMPP bots for humans")
+
         self.parser.add_argument(
             "-d",
             "--debug",
-            help="Set logging to DEBUG",
+            help="Enable verbose debug logs",
             action="store_const",
             dest="log_level",
             const=DEBUG,
             default=INFO,
         )
+        self.parser.add_argument(
+            "-a",
+            "--account",
+            dest="account",
+            help="Account for the bot account (foo@example.com)",
+        )
+        self.parser.add_argument(
+            "-p",
+            "--password",
+            dest="password",
+            help="Password for the bot account",
+        )
+        self.parser.add_argument(
+            "-n",
+            "--nick",
+            dest="nick",
+            help="Nickname for the bot account",
+        )
+
         self.args = self.parser.parse_args()
 
     def setup_logging(self):
@@ -79,53 +122,68 @@ class Bot(ClientXMPP):
         basicConfig(
             level=self.args.log_level, format="%(levelname)-8s %(message)s"
         )
+        self.log = getLogger(__name__)
 
     def read_config(self):
         """Read configuration for running bot."""
-        self.config = ConfigParser()
+        config = ConfigParser()
 
         config_file_path = Path(self.CONFIG_FILE).absolute()
-        if not exists(config_file_path) and self.args.input:
+
+        if not exists(config_file_path) and stdout.isatty():
             self.generate_config_interactively()
 
         if exists(config_file_path):
-            self.config.read(config_file_path)
+            config.read(config_file_path)
 
-        if self.args.input is False:
-            self.read_config_from_env()
-
-        if "bot" not in self.config:
-            raise RuntimeError("Failed to configure bot")
+        self.config = Config(self.name, config)
 
     def generate_config_interactively(self):
         """Generate bot configuration."""
-        jid = input("XMPP address (e.g. foo@bar.com): ") or "foo@bar.com"
-        password = (
-            getpass("Password (e.g. my-cool-password): ") or "my-cool-password"
-        )
-        nick = input("Nickname (e.g. lurkbot): ")
+        account = input("Account: ")
+        password = getpass("Password: ")
+        nick = input("Nickname: ")
 
         config = ConfigParser()
-        config["bot"] = {"jid": jid, "password": password}
+        config[self.name] = {"account": account, "password": password}
 
         if nick:
-            config["bot"]["nick"] = nick
+            config[self.name]["nick"] = nick
 
-        with open("bot.conf", "w") as file_handle:
+        with open(self.CONFIG_FILE, "w") as file_handle:
             config.write(file_handle)
 
-    def read_config_from_env(self):
-        """Read configuration from the environment."""
-        self.config["bot"] = {}
-        self.config["bot"]["jid"] = environ.get("XBOT_JID")
-        self.config["bot"]["password"] = environ.get("XBOT_PASSWORD")
-        self.config["bot"]["nick"] = environ.get("XBOT_NICK", "")
-
     def init_bot(self):
         """Initialise bot with connection details."""
-        jid = self.config["bot"]["jid"]
-        password = self.config["bot"]["password"]
-        ClientXMPP.__init__(self, jid, password)
+        account = (
+            self.args.account
+            or self.config.account
+            or environ.get("XBOT_ACCOUNT", None)
+        )
+        password = (
+            self.args.password
+            or self.config.password
+            or environ.get("XBOT_PASSWORD", None)
+        )
+        nick = (
+            self.args.nick or self.config.nick or environ.get("XBOT_NICK", None)
+        )
+
+        if not account:
+            self.log.error("Unable to discover account")
+            exit(1)
+        if not password:
+            self.log.error("Unable to discover password")
+            exit(1)
+        if not nick:
+            self.log.error("Unable to discover nick")
+            exit(1)
+
+        ClientXMPP.__init__(self, account, password)
+
+        self.account = account
+        self.password = password
+        self.nick = nick
 
     def register_xmpp_event_handlers(self):
         """Register functions against specific XMPP event handlers."""
@@ -137,7 +195,16 @@ class Bot(ClientXMPP):
     def direct_message(self, message):
         """Handle message event."""
         if message["type"] in ("chat", "normal"):
-            self.direct(SimpleMessage(message))
+            _message = SimpleMessage(message)
+
+            if _message.body.startswith("!"):
+                self.command(_message, to=_message.sender)
+                return
+
+            try:
+                self.direct(_message)
+            except AttributeError:
+                self.log.info("Bot.direct not implemented")
 
     def session_start(self, message):
         """Handle session_start event."""
@@ -146,15 +213,22 @@ class Bot(ClientXMPP):
 
     def group_invite(self, message):
         """Accept invites to group chats."""
-        self.plugin["xep_0045"].join_muc(
-            message["from"], self.config["bot"]["nick"]
-        )
+        self.plugin["xep_0045"].join_muc(message["from"], self.config.nick)
 
     def group_message(self, message):
         """Handle groupchat_message event."""
         if message["type"] in ("groupchat", "normal"):
-            if message["mucnick"] != self.config["bot"]["nick"]:
-                self.group(SimpleMessage(message))
+            if message["mucnick"] != self.config.nick:
+                _message = SimpleMessage(message)
+
+                if f"{self.nick}:!" in _message.body:
+                    self.command(_message, room=_message.room)
+                    return
+
+                try:
+                    self.group(_message)
+                except AttributeError:
+                    self.log.info("Bot.group not implemented")
 
     def register_xmpp_plugins(self):
         """Register XMPP plugins that the bot supports."""
@@ -174,12 +248,12 @@ class Bot(ClientXMPP):
     def reply(self, body, to=None, room=None):
         """Send back a reply."""
         if to is None and room is None:
-            message = "`to` or `room` arguments required for `reply`"
-            raise RuntimeError(message)
+            self.log("`to` or `room` arguments required for `reply`")
+            exit(1)
 
         if to is not None and room is not None:
-            message = "Cannot send to both `to` and `room` for `reply`"
-            raise RuntimeError(message)
+            self.log.error("Cannot send to both `to` and `room` for `reply`")
+            exit(1)
 
         kwargs = {"mbody": body}
         if to is not None:
@@ -191,9 +265,24 @@ class Bot(ClientXMPP):
 
         self.send_message(**kwargs)
 
-    def react(self, message):
-        message = "You need to write your own `react` implementation"
-        raise NotImplementedError(message)
+    @property
+    def uptime(self):
+        """Time since the bot came up."""
+        return naturaldelta(self.start - dt.now())
+
+    def command(self, message, **kwargs):
+        """Handle "!" style commands with built-in responses."""
+        command = message.body.split("!")[-1]
+
+        if command == "uptime":
+            self.reply(self.uptime, **kwargs)
+        elif command == "help":
+            try:
+                self.reply(self.help, **kwargs)
+            except AttributeError:
+                self.reply("No help found 🤔️", **kwargs)
+        else:
+            self.log.error(f"'{command}' direct command is not recognised")
 
 
 class EchoBot(Bot):
@@ -207,6 +296,8 @@ class EchoBot(Bot):
 
     """
 
+    help = "I echo back whatever you send to me 🖖️"
+
     def direct(self, message):
         """Send back whatever we receive."""
         self.reply(message.body, to=message.sender)
@@ -228,6 +319,8 @@ class WhisperBot(Bot):
 
     """
 
+    help = "I whisper your private messages into group chats 😌️"
+
     def direct(self, message):
         """Receive private messages and whisper them into group chats."""
         self.reply(f"*pssttt...* {message.body}", room=message.room)
